"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Delete elements from model.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from Autodesk.Revit.DB import Element, ElementId, Transaction

from duHast.Revit.Common import transaction as rTran
from duHast.Utilities.Objects import result as res

# required for .ToList()
import clr

clr.AddReference("System.Core")
from System import Linq

clr.ImportExtensions(Linq)


def delete_by_element_ids(
    doc,
    ids,
    transaction_name,  # type: str
    element_name,  # type: str
):
    """
    Deletes elements in a list all at once.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param ids: List containing IDs of all elements to be deleted.
    :type ids: list of Autodesk.Revit.DB.ElementId
    :param transaction_name: The transaction name used for the deletion.
    :type transaction_name: str
    :param element_name: The element name added to the deletion status message.
    :type element_name: str

    :return: Result class instance.
        - `result` (bool): True if all elements were successfully deleted, otherwise False.
        - `message` (str): Deletion status.
    :rtype: :class:`.Result`
    """

    return_value = res.Result()

    def action():
        action_return_value = res.Result()
        try:
            doc.Delete(ids.ToList[ElementId]())
            action_return_value.message = "Deleted {} {}".format(len(ids), element_name)
        except Exception as e:
            action_return_value.update_sep(
                False, "Failed to delete {} with exception: {}".format(element_name, e)
            )
        return action_return_value

    transaction = Transaction(doc, transaction_name)
    return_value = rTran.in_transaction(transaction, action)
    return return_value


def delete_by_element_ids_one_by_one(
    doc,
    ids,
    transaction_name,  # type: str
    element_name,  # type: str
):
    """
    Deletes elements in a list one at a time.

    Each element is deleted in its own transaction. If the deletion fails, the transaction is rolled back.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param ids: List containing IDs of all elements to be deleted.
    :type ids: list of Autodesk.Revit.DB.ElementId
    :param transaction_name: The transaction name used for the deletion.
    :type transaction_name: str
    :param element_name: The name of the element (not used).
    :type element_name: str

    :return: Result class instance.
        - `result` (bool): True if all elements were successfully deleted, otherwise False.
        - `message` (str): Contains each ID and its deletion status.
    :rtype: :class:`.Result`

    """

    return_value = res.Result()
    for id in ids:

        def action():
            action_return_value = res.Result()
            element = doc.GetElement(id)
            n = Element.Name.GetValue(element)
            try:
                doc.Delete(id)
                action_return_value.message = "Deleted [{}] {}".format(id, n)
            except Exception as e:
                action_return_value.update_sep(
                    False,
                    "Failed to delete {} [{}] with exception: {}".format(n, id, e),
                )
            return action_return_value

        transaction = Transaction(doc, transaction_name)
        return_value.update(rTran.in_transaction(transaction, action))
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Family shared parameter data class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.Family.Data.Objects import ifamily_data as IFamData
from duHast.Utilities import utility as util
from duHast.Revit.SharedParameters import shared_parameters as rSharedPara

# import Autodesk
import Autodesk.Revit.DB as rdb


PARAMETER_GUID = "parameterGUID"
PARAMETER_NAME = "parameterName"
PARAMETER_ID = "parameterId"


class SharedParameterData(IFamData.IFamilyData):
    def __init__(self, root_path=None, root_category_path=None, data_type=None):

        super(SharedParameterData, self).__init__(
            root_path=root_path,
            root_category_path=root_category_path,
            data_type=data_type,
        )
        # super(CategoryData, self).__init__(rootPath, dataType)

        """
        self.data = []
        
        if(dataType != None):
            self.dataType = dataType
        else:
            self.dataType = 'not declared'
        
        if(rootPath != None):
            self.rootPath = rootPath
        else:
            self.rootPath = '-'

        if(rootCategoryPath != None):
            self.rootCategoryPath = rootCategoryPath
        else:
            self.rootCategoryPath = '-'
        """

    def process(self, doc):
        collector = rSharedPara.get_all_shared_parameters(doc)
        for para in collector:
            # just in case parameter name is not unicode
            parameter_name = "unknown"
            try:
                parameter_name = util.encode_ascii(rdb.Element.Name.GetValue(para))
            except Exception as ex:
                parameter_name = "Exception: " + str(ex)
            # check if used:
            use_counter = 0
            used_by_data = {}
            if rSharedPara.is_shared_parameter_definition_used(doc, para):
                use_counter = 1
                # build used by data as required to be the same as post process update
                used_by_data = {
                    PARAMETER_GUID: para.GuidValue.ToString(),
                    PARAMETER_NAME: parameter_name,
                    IFamData.ROOT: self.root_path,
                }

            # build data
            self.data.append(
                {
                    IFamData.ROOT: self.root_path,
                    IFamData.ROOT_CATEGORY: self.root_category_path,
                    IFamData.FAMILY_NAME: self._strip_file_extension(doc.Title),
                    IFamData.FAMILY_FILE_PATH: doc.PathName,
                    PARAMETER_GUID: para.GuidValue.ToString(),
                    PARAMETER_NAME: parameter_name,
                    PARAMETER_ID: para.Id.IntegerValue,
                    IFamData.USAGE_COUNTER: use_counter,
                    IFamData.USED_BY: [used_by_data],
                }
            )

        # check if any shared parameter was found
        if len(self.data) == 0:
            # add message no shared parameter found
            self.data.append(
                {
                    IFamData.ROOT: self.root_path,
                    IFamData.ROOT_CATEGORY: self.root_category_path,
                    IFamData.FAMILY_NAME: self._strip_file_extension(doc.Title),
                    IFamData.FAMILY_FILE_PATH: doc.PathName,
                    PARAMETER_GUID: "",
                    PARAMETER_NAME: "No shared parameter present in family.",
                    PARAMETER_ID: -1,
                    IFamData.USAGE_COUNTER: 0,
                    IFamData.USED_BY: [],
                }
            )

    def get_data(self):
        return self.data



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit areas helper functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from Autodesk.Revit.DB import (
    BuiltInCategory,
    ElementClassFilter,
    FilteredElementCollector,
    ViewPlan
)


def get_areas_by_scheme_name(doc, scheme_name):
    """
    Retrieves a list of areas that belong to a specific area scheme.

    Args:
        doc (Revit document): The Revit document from which to retrieve the areas.
        scheme_name (str): The name of the area scheme to filter the areas by.

    Returns:
        list: A list of areas that belong to the specified area scheme.
    """
    return_value = []
    areas = FilteredElementCollector(doc).OfCategory(BuiltInCategory.OST_Areas)
    for area in areas:
        if area.AreaScheme.Name == scheme_name:
            return_value.append(area)
    return return_value


def get_area_schemes(doc):
    """
    Retrieves all area schemes in Autodesk Revit.

    Args:
        doc (Revit Document): The Revit document in which to search for area schemes.

    Returns:
        area_schemes: A collection of area schemes in the Revit document.
    """
    area_schemes = FilteredElementCollector(doc).OfCategory(
        BuiltInCategory.OST_AreaSchemes
    )
    return area_schemes


def get_area_scheme_by_name(doc, area_scheme_name):
    """
    Retrieves an area scheme by its name in Autodesk Revit.

    Args:
        doc (Revit Document): The Revit document in which to search for the area scheme.
        area_scheme_name (str): The name of the area scheme to retrieve.

    Returns:
        AreaScheme: The area scheme with the specified name, or None if no match is found.
    """
    return_value = None
    area_schemes = get_area_schemes(doc=doc)
    if any(area_schemes):
        for area_scheme in area_schemes:
            if area_scheme.Name == area_scheme_name:
                return area_scheme
    return return_value

def get_views_by_area_scheme_name(doc, area_scheme_name):
    """
    Retrieves all views associated with a given area scheme name in Autodesk Revit.

    Args:
        doc (Revit Document): The Revit document in which to search for views.
        area_scheme_name (str): The name of the area scheme to filter views by.

    Returns:
        list: A list of ViewPlan objects that are associated with the given area scheme name.
    """
    return_value = []
    area_scheme = get_area_scheme_by_name(doc, area_scheme_name)
    if area_scheme:
        filter = ElementClassFilter(ViewPlan)
        dependent_element_ids = area_scheme.GetDependentElements(filter)
        for id in dependent_element_ids:
            element = doc.GetElement(id)
            return_value.append(element)
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit transaction wrapper utility functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#!/usr/bin/python
# -*- coding: utf-8 -*-
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#

# class used for stats reporting
from duHast.Utilities.Objects import result as res
from duHast.Revit.Common.Objects.FailuresPreProcessor import FailuresPreprocessor
from duHast.Revit.Common.failure_handling import process_failures
from duHast.Revit.Common.Objects.FailureHandlingConfiguration import (
    FailureHandlingConfig,
)

from Autodesk.Revit.DB import Transaction

# --------------------------------------------Transactions-----------------------------------------


def in_transaction(tranny, action, *args, **kwargs):
    """
    Revit transaction wrapper.

    This function is used to execute any actions requiring a transaction in the Revit api. On exception this will roll back the transaction.

    :param tranny: The transaction to be executed.
    :type tranny: Autodesk.Revit.DB.Transaction
    :param doc: The current model document. (not used in this function)
    :type doc: Autodesk.Revit.DB.Document
    :param action: The action to be nested within the transaction. This needs to return a Result class instance!
    :type action: action().
    :param *args: is just a placeholder in case this function is called with the same args than in_transaction_with_failure_handling
    :param **kwargs: is just a placeholder in case this function is called with the same args than in_transaction_with_failure_handling

    :return:
        Result class instance.

        - .status True if successfully executed transaction, otherwise False.

    :rtype: :class:`.Result`
    """

    if not isinstance(tranny, Transaction):
        raise ValueError(
            "The transaction parameter must be an instance of Autodesk.Revit.DB.Transaction."
        )
    if not callable(action):
        raise ValueError("The action parameter must be a callable function.")

    return_value = res.Result()
    try:
        tranny.Start()
        try:
            tranny_result = action()
            tranny.Commit()
            # check what came back
            if tranny_result != None:
                # store false value
                return_value = tranny_result
        except Exception as e:
            tranny.RollBack()
            return_value.update_sep(False, "Failed with exception: {}".format(e))
    except Exception as e:
        return_value.update_sep(False, "Failed with exception: {}".format(e))
   
    return return_value


def in_transaction_with_failure_handling(
    transaction,
    action,
    failure_config=FailureHandlingConfig(),
    failure_processing_func=process_failures,
):
    """
    Executes an action within a Revit transaction with Revit failure processing enabled.

    Example usage:

    def main(revit_doc, foo, bar):

        fail_config = FailureHandlingConfig()
        fail_config.print_warnings = False

        def add_func():
            # Perform actions here
            return foo + bar

        trans = Transaction(revit_doc, "Add the things together")
        added_vals = in_transaction_with_failures(trans, add_func, fail_config)

    :param transaction: The executing transaction to apply the failure handling to
    :type transaction: Transaction
    :param action: The action to be executed within the transaction
    :type action: function
    :param failure_config: Configuration for the failure handling (Optional)
    :type failure_config: FailureHandlingConfig
    :param failure_processing_func: The function to process the failures (Optional)
    :type failure_processing_func: function

    """


    if not callable(action):
        raise ValueError("The action parameter must be a callable function.")
    if not callable(failure_processing_func):
        raise ValueError(
            "The failure_processing_func parameter must be a callable function."
        )
    if not isinstance(failure_config, FailureHandlingConfig):
        raise ValueError(
            "The failure_config parameter must be an instance of FailureHandlingConfig."
        )
    if not isinstance(transaction, Transaction):
        raise ValueError(
            "The transaction parameter must be an instance of Autodesk.Revit.DB.Transaction."
        )

    # Establish the failure preprocessor
    failure_pre_processor = FailuresPreprocessor(
        failure_processor=failure_processing_func, 
        fail_config=failure_config
    )
    # Get and overwrite the failure handling options for the transaction arg
    failure_opts = transaction.GetFailureHandlingOptions()
    failure_opts.SetForcedModalHandling(failure_config.set_forced_modal_handling)
    failure_opts.SetClearAfterRollback(failure_config.set_clear_after_rollback)
    failure_opts.SetFailuresPreprocessor(failure_pre_processor)
    transaction.SetFailureHandlingOptions(failure_opts)

    # Execute the action with the updated transaction
    result = in_transaction(transaction, action)
    return result



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit sub-category property get functions .
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


import Autodesk.Revit.DB as rdb
from duHast.Revit.LinePattern import line_patterns as rPat
from duHast.Revit.Categories.Utility.category_property_names import (
    CATEGORY_GRAPHIC_STYLE_3D,
    CATEGORY_GRAPHIC_STYLE_CUT,
    CATEGORY_GRAPHIC_STYLE_PROJECTION,
    PROPERTY_LINE_COLOUR_BLUE_NAME,
    PROPERTY_LINE_COLOUR_GREEN_NAME,
    PROPERTY_LINE_COLOUR_RED_NAME,
    PROPERTY_LINE_WEIGHT_CUT_NAME,
    PROPERTY_LINE_WEIGHT_PROJECTION_NAME,
    PROPERTY_MATERIAL_ID,
    PROPERTY_MATERIAL_NAME,
    PROPERTY_MATERIAL_NAME_VALUE_DEFAULT,
)


def get_category_graphic_style_ids(cat):
    """
    Returns a dictionary with keys: Projection, Cut, 3D and their respective ids
    :param cat: A category.
    :type cat: Autodesk.REvit.DB.Category
    :return: A dictionary
    :rtype: dictionary {str: Autodesk.Revit.DB.ElementId}
    """

    i_d_graphic_style_projection = cat.GetGraphicsStyle(
        rdb.GraphicsStyleType.Projection
    ).Id

    # check if this category has a cut style ( some families always appear in elevation only!)
    graphic_style_cut = cat.GetGraphicsStyle(rdb.GraphicsStyleType.Cut)
    # set as default invalid element id
    i_d_graphic_style_cut = rdb.ElementId.InvalidElementId
    if graphic_style_cut != None:
        i_d_graphic_style_cut = cat.GetGraphicsStyle(rdb.GraphicsStyleType.Cut).Id
    # build category dictionary where key is the style type, values is the corresponding Id
    dic = {}
    dic[CATEGORY_GRAPHIC_STYLE_PROJECTION] = i_d_graphic_style_projection
    dic[CATEGORY_GRAPHIC_STYLE_CUT] = i_d_graphic_style_cut
    dic[CATEGORY_GRAPHIC_STYLE_3D] = cat.Id
    return dic


def get_category_material(cat):
    """
    Returns the material properties name and id as a dictionary where key is property name and\
         value the property id.
    :param cat: A category.
    :type cat: Autodesk.REvit.DB.Category
    :return: A dictionary
    :rtype: dictionary {str: Autodesk.Revit.DB.ElementId}\
        If no material is assigned to a category it will return {'None: Autodesk.Revit.DB.ElementId.InvalidElementId}
    """

    dic_material = {}
    dic_material[PROPERTY_MATERIAL_NAME] = PROPERTY_MATERIAL_NAME_VALUE_DEFAULT
    dic_material[PROPERTY_MATERIAL_ID] = rdb.ElementId.InvalidElementId
    material = cat.Material
    if material != None:
        dic_material[PROPERTY_MATERIAL_NAME] = rdb.Element.Name.GetValue(material)
        dic_material[PROPERTY_MATERIAL_ID] = material.Id
    return dic_material


def get_category_line_weights(cat):
    """
    Returns the line weight properties (cut and projection) as a dictionary\
         where key is property description and value the property value
    :param cat: A category.
    :type cat: Autodesk.Revit.DB.Category
    :return: A dictionary.
    :rtype: dictionary {str: nullable integer}
    """

    dic_line_weights = {}
    dic_line_weights[PROPERTY_LINE_WEIGHT_PROJECTION_NAME] = cat.GetLineWeight(
        rdb.GraphicsStyleType.Projection
    )
    dic_line_weights[PROPERTY_LINE_WEIGHT_CUT_NAME] = cat.GetLineWeight(
        rdb.GraphicsStyleType.Cut
    )
    return dic_line_weights


def get_category_colour(cat):
    """
    Returns the colour properties (RGB) and values as a dictionary where key is colour name\
         and value the property value
    :param cat: A category.
    :type cat: Autodesk.Revit.DB.Category
    :return: A dictionary.
    :rtype: dictionary {str: byte}
    """

    dic_colour = {}
    dic_colour[PROPERTY_LINE_COLOUR_RED_NAME] = 0
    dic_colour[PROPERTY_LINE_COLOUR_GREEN_NAME] = 0
    dic_colour[PROPERTY_LINE_COLOUR_BLUE_NAME] = 0
    if cat.LineColor.IsValid:
        dic_colour[PROPERTY_LINE_COLOUR_RED_NAME] = cat.LineColor.Red
        dic_colour[PROPERTY_LINE_COLOUR_GREEN_NAME] = cat.LineColor.Green
        dic_colour[PROPERTY_LINE_COLOUR_BLUE_NAME] = cat.LineColor.Blue
    return dic_colour


def get_category_properties(cat, doc):
    """
    Returns a dictionary where keys are category property names and value is the associated property value.
    :param cat: A category.
    :type cat: Autodesk.Revit.DB.Category
    :param doc: Current Revit family document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A dictionary.
    :rtype: list [{str: var}]
    """

    properties = []

    # material
    dic_material = get_category_material(cat)
    properties.append(dic_material)

    # line pattern
    dic_pattern = rPat.get_line_pattern_from_category(cat, doc)
    properties.append(dic_pattern)

    # line weights
    dic_line_weights = get_category_line_weights(cat)
    properties.append(dic_line_weights)

    # category colour
    dic_colour = get_category_colour(cat)
    properties.append(dic_colour)
    return properties


def get_saved_category_property_by_name(properties, prop_names):
    """
    Returns property values matching property names in saved category data.
    :param properties: List of dictionaries in format as per GetCategoryProperties(cat) method.
    :type properties: list [{str: var}]
    :param prop_names: List of property names of which the values are to be returned
    :type prop_names: list str
    :return: A list of values.
    :rtype: list var
    """

    prop_values = []
    for prop_name in prop_names:
        match = False
        for saved_prop in properties:
            if prop_name in saved_prop:
                prop_values.append(saved_prop[prop_name])
                match = True
        if match == False:
            prop_values.append(None)
    return prop_values



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to CAD link geometry.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import Autodesk.Revit.DB as rdb

from duHast.Revit.Links.cad_links import get_all_cad_link_instances


def get_cad_import_instance_geometry(import_instance):
    """
    Returns a list of geometry elements from an import instance
    :param import_instance: A import instance
    :type import_instance: AutoDesk.Revit.DB.ImportInstance
    :return: A list of geometry objects. Can return an empty list!
    :rtype: [Autodesk.Revit.DB GeometryObject]
    """

    geo = []
    # default geometry option
    opt = rdb.Options()
    geo_elem_level1 = import_instance.get_Geometry(opt)
    if geo_elem_level1 != None:
        for geo_instance in geo_elem_level1:
            if geo_instance != None:
                geo_elem_level2 = geo_instance.GetInstanceGeometry()
                if geo_elem_level2 != None:
                    for item in geo_elem_level2:
                        geo.append(item)
    return geo


def get_all_cad_import_instances_geometry(doc):
    """
    Returns a list of geometry elements from all import instances in the document.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A list of geometry objects. Can return an empty list!
    :rtype: [Autodesk.Revit.DB GeometryObject]
    """
    instances_geometry = []
    all_import_instances = get_all_cad_link_instances(doc)
    for import_instance in all_import_instances:
        geometry_instances = get_cad_import_instance_geometry(import_instance)
        if len(geometry_instances) > 0:
            instances_geometry += geometry_instances
    return instances_geometry



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This class is the default implementation for family load call backs.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

import clr

from Autodesk.Revit.DB import IFamilyLoadOptions, FamilySource


class FamilyLoadOption(IFamilyLoadOptions):
    def OnFamilyFound(self, familyInUse, overwriteParameterValues):
        """
        Defines behavior when a family is found in the model.

        Overwrite parameter values is set to True.

        :param familyInUse: _description_
        :type familyInUse: _type_
        :param overwriteParameterValues: _description_
        :type overwriteParameterValues: _type_
        :return: True
        :rtype: bool
        """
        overwriteParameterValues = True
        return True

    def OnSharedFamilyFound(
        self, sharedFamily, familyInUse, source, overwriteParameterValues
    ):
        """
        Defines behavior when a shared family is found in the model.

        Overwrite parameter values is set to True. In case of any shared sub components already in the model but different in family loaded, the project version will be used.

        :param sharedFamily: _description_
        :type sharedFamily: _type_
        :param familyInUse: _description_
        :type familyInUse: _type_
        :param source: _description_
        :type source: _type_
        :param overwriteParameterValues: _description_
        :type overwriteParameterValues: _type_
        :return: True
        :rtype: bool
        """
        source = FamilySource.Project
        overwriteParameterValues = True
        return True



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains the Revit sheet report functionality.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from Autodesk.Revit.DB import FilteredElementCollector, ViewSheet, WorksharingUtils

from duHast.Revit.Views.Reporting.views_report_header import (
    REPORT_SHEETS_HEADER,
    get_sheets_report_headers,
)
from duHast.Revit.Views.Reporting.view_property_filter import filter_data_by_properties
from duHast.Revit.Common import parameter_get_utils as rParaGet
from duHast.Utilities.Objects import result as res
from duHast.Revit.Views.Reporting.view_property_utils import convert_view_data_to_list
from duHast.Utilities.files_csv import write_report_data_as_csv


def get_sheet_report_data(doc, host_name):
    """
    Gets sheet data to be written to report file.
    The data returned includes all sheet properties available in the file.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param host_name: The file hostname, which is added to data returned
    :type host_name: str
    :return: list of list of sheet properties.
    :rtype: list of list of str
    """

    collector_views = FilteredElementCollector(doc).OfClass(ViewSheet)
    views = []
    for v in collector_views:
        # get all parameters attached to sheet
        paras = v.GetOrderedParameters()
        info = WorksharingUtils.GetWorksharingTooltipInfo(doc, v.Id)
        # check values and substitute any empty strings
        owner = info.Owner
        last_changed_by = (info.LastChangedBy,)
        if info.Owner == "":
            owner = "not applicable"
        if info.LastChangedBy == "":
            last_changed_by = "not applicable"
        data = {
            REPORT_SHEETS_HEADER[0]: host_name,
            REPORT_SHEETS_HEADER[1]: str(v.Id),
            REPORT_SHEETS_HEADER[2]: info.Creator,
            REPORT_SHEETS_HEADER[3]: last_changed_by,
            REPORT_SHEETS_HEADER[4]: owner,
        }
        for para in paras:
            # get values as utf-8 encoded strings
            value = rParaGet.get_parameter_value_utf8_string(para)
            try:
                data[para.Definition.Name] = value
            except:
                data[para.Definition.Name] = "Failed to retrieve value"
        views.append(data)
    return views


def get_sheets_report_data_filtered(doc, host_name, sheet_properties):
    """
    Gets sheet data to be written to report file.
    The data returned is filtered by property list past in.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param host_name: The file hostname, which is added to data returned
    :type host_name: str
    :param view_properties: List of view properties to be extracted from sheets.
    :type view_properties: list of str
    :return: list of list of view properties.
    :rtype: list of list of str
    """

    data = get_sheet_report_data(doc, host_name)
    headers = get_sheets_report_headers(doc)
    sheets_filtered = filter_data_by_properties(
        data, headers, sheet_properties, REPORT_SHEETS_HEADER
    )
    return sheets_filtered


def write_sheet_data(doc, file_name, current_file_name):
    """
    Writes to file all sheet properties.

    file type: csv

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param file_name: The fully qualified file path of the report file.
    :type file_name: str
    :param current_file_name: The current revit file name which will be appended to data in the report.
    :type current_file_name: str
    :return:
        Result class instance.
        - .status True if data was written successfully. Otherwise False.
        - .message will contain write status.
    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    try:
        data = get_sheet_report_data(doc, current_file_name)
        headers = get_sheets_report_headers(doc)
        data_converted = convert_view_data_to_list(data, headers)
        write_report_data_as_csv(file_name, headers, data_converted)
        return_value.update_sep(True, "Successfully wrote data file")
    except Exception as e:
        return_value.update_sep(False, str(e))
    return return_value


def write_sheet_data_by_property_names(
    doc, file_name, current_file_name, sheet_properties
):
    """
    Writes to file sheet properties as nominated in past in list.

    file type: csv

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param file_name: The fully qualified file path of the report file.
    :type file_name: str
    :param current_file_name: The current Revit file name which will be appended to data in the report.
    :type current_file_name: str
    :param sheet_properties: List of sheet properties to be extracted from sheets.
    :type sheet_properties: list of str
    :return:
        Result class instance.
        - .status True if data was written successfully. Otherwise False.
        - .message will contain write status.
    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    try:
        data = get_sheets_report_data_filtered(doc, current_file_name, sheet_properties)
        # change headers to filtered + default
        headers = REPORT_SHEETS_HEADER[:] + sheet_properties
        data_converted = convert_view_data_to_list(data, headers)
        # write data out to file
        write_report_data_as_csv(file_name, headers, data_converted)
        return_value.update_sep(True, "Successfully wrote data file")
    except Exception as e:
        return_value.update_sep(False, str(e))
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Finds host families of nested families requiring to be renamed.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

- read file rename change list:
-   text file (tab separated) with columns: currentFamilyName   currentFamilyFilePath	categoryName newFamilyName
-   note: current family file path could be blank in situations where just a rename of a nested family is required...
- read base data file processing list (created by RevitFamilyBaseDataProcessor module)
- extract all root families (no :: in root path)
- extract all nested families ( :: in root path)
- loop over nested families and find any family where:
    - family at first nesting level is in rename list by name and category
    - extract root family name and add to identified host family list

- loop over root families
-   find match in identified host families
- write out root family date: family file path, family name, category

"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

# import clr
# import System
from duHast.Utilities.Objects.timer import Timer

from duHast.Utilities.Objects import result as res

# from duHast.Utilities import Utility as util
from duHast.Revit.Family.Data import family_base_data_utils as rFamBaseDataUtils
from duHast.Revit.Family import family_rename_files_utils as rFamRenameUtils

# ---------------------------------------------------------------------------------------------------------------
#                      find families containing nested families needing to be renamed
# ---------------------------------------------------------------------------------------------------------------


def _find_host_families(overall_family_base_nested_data, file_rename_list):
    """
    Finds all root family names and categories where the first level nested family is one which needs to be renamed.

    :param overall_family_base_nested_data: A list containing all root families.
    :type overall_family_base_nested_data: [rootFamily]
    :param file_rename_list: A list containing all families needing to be renamed.
    :type file_rename_list: [renameFamily]

    :return: A dictionary where:

        - key is the family name and category concatenated and
        - value is a tuple in format 0: family name, 1: family category

    :rtype: {str: (str,str)}
    """

    host_families = {}
    for file_rename_family in file_rename_list:
        hosts = rFamBaseDataUtils.find_direct_host_families(
            file_rename_family, overall_family_base_nested_data
        )
        # update dictionary with new hosts only
        for h in hosts:
            if h not in host_families:
                host_families[h] = hosts[h]
    return host_families


def find_host_families_with_nested_families_requiring_rename(input_directory_path):
    """
    Finds all host families in data set containing nested families needing to be renamed.

    :param input_directory_path: Fully qualified directory path containing rename directives and family base data report.
    :type input_directory_path: str

    :return:
        Result class instance.

        - result.status: True if any host families are found with nested families needing renaming, otherwise False.
        - result.message: processing steps with time stamps
        - result.result: [rootFamily]

        On exception:

        - result.status (bool) will be False.
        - result.message will contain an exception message
        - result.result will be empty

    :rtype: :class:`.Result`
    """

    # set up a timer
    t_process = Timer()
    t_process.start()

    return_value = res.Result()
    # read overall family base data from file
    try:
        (
            overall_family_base_root_data,
            overall_family_base_nested_data,
        ) = rFamBaseDataUtils.read_overall_family_data_list_from_directory(
            input_directory_path
        )
        return_value.append_message(
            "{} Read overall family base data report. {} root entries found and {} nested entries found.".format(
                t_process.stop(),
                len(overall_family_base_root_data),
                len(overall_family_base_nested_data),
            )
        )
        # check if input file existed and contained data
        if len(overall_family_base_root_data) > 0:
            t_process.start()
            file_rename_list_status = rFamRenameUtils.get_rename_directives(
                input_directory_path
            )
            return_value.append_message(
                "{} Read data from file! Rename family entries [{}] found.".format(
                    t_process.stop(), len(file_rename_list_status.result)
                )
            )
            # check if any rename directives
            if len(file_rename_list_status.result) > 0:
                before = len(overall_family_base_nested_data)
                t_process.start()
                # reduce workload by culling not needed nested family data
                overall_family_base_nested_data = (
                    rFamBaseDataUtils.cull_nested_base_data_blocks(
                        overall_family_base_root_data, overall_family_base_nested_data
                    )
                )
                return_value.append_message(
                    "{} Culled nested family base data from : {} to: {} families.".format(
                        t_process.stop(), before, len(overall_family_base_nested_data)
                    )
                )

                t_process.start()
                # get a list of simplified root data families extracted from nested family path data
                root_fam_simple = rFamBaseDataUtils.find_all_direct_host_families(
                    file_rename_list_status.result, overall_family_base_nested_data
                )
                return_value.append_message(
                    "{} Found simplified root families: {}".format(
                        t_process.stop(), len(root_fam_simple)
                    )
                )

                t_process.start()
                # identify actual root families with nested families at top level which require renaming.
                root_families = rFamBaseDataUtils.find_root_families_from_hosts(
                    root_fam_simple, overall_family_base_root_data
                )
                return_value.append_message(
                    "{} Found {} root families.".format(
                        t_process.stop(), len(root_families)
                    )
                )
                return_value.result = root_families
            else:
                return_value.update_sep(
                    False, "No rename directives found. Aborted operation!"
                )
        else:
            return_value.update_sep(
                False, "No base family data found. Aborted operation!"
            )
    except Exception as e:
        return_value.update_sep(
            False, "Failed to find host families with exception: {}".format(e)
        )
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to purging unused families which can be used in mep systems.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


# -------------------------------- purge loaded families which can be used in system types --------------------------------

from duHast.Revit.Family import purge_unused_family_types as rFamPurge
from duHast.Revit.Common import purge_utils as rPurgeUtils

from duHast.Revit.MEP_Systems.pipes import (
    get_symbol_ids_for_pipe_types_in_model,
    get_symbol_ids_used_in_pipe_types,
)
from duHast.Revit.MEP_Systems.conduits import (
    get_symbol_ids_for_conduit_types_in_model,
    get_symbol_ids_used_in_conduit_types,
)
from duHast.Revit.MEP_Systems.cable_trays import (
    get_symbol_ids_for_cable_tray_types_in_model,
    get_symbol_ids_used_in_cable_tray_types,
)
from duHast.Revit.MEP_Systems.ducts import (
    get_symbol_ids_for_duct_types_in_model,
    get_symbol_ids_used_in_duct_types,
)
from duHast.Revit.MEP_Systems.flex_ducts import get_symbol_ids_used_in_flex_duct_types
from duHast.Revit.MEP_Systems.Utility.MergeLists import merge_into_unique_list


def get_used_duct_and_flex_duct_symbol_ids(doc):
    """
    Gets all used duct and flex duct symbol ids of categories
    - BuiltInCategory.OST_DuctAccessory,
    - BuiltInCategory.OST_DuctTerminal,
    - BuiltInCategory.OST_DuctFitting
    Used: at least instance per symbol is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of ids representing family symbols.
    :rtype: list  Autodesk.Revit.DB.ElementId
    """

    ids = []
    ids_in_model = rPurgeUtils.get_used_unused_type_ids(
        doc, get_symbol_ids_for_duct_types_in_model, 1
    )
    ids_used_in_types = get_symbol_ids_used_in_duct_types(doc)
    ids_used_in_flex_types = get_symbol_ids_used_in_flex_duct_types(doc)
    ids = merge_into_unique_list(ids, ids_in_model)
    ids = merge_into_unique_list(ids, ids_used_in_types)
    ids = merge_into_unique_list(ids, ids_used_in_flex_types)
    return ids


def get_unused_duct_and_flex_duct_symbol_ids(doc):
    """
    Gets all unused duct and flex duct symbol ids of categories
    - BuiltInCategory.OST_DuctAccessory,
    - BuiltInCategory.OST_DuctTerminal,
    - BuiltInCategory.OST_DuctFitting
    Unused: not one instance per symbol is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of ids representing family symbols.
    :rtype: list  Autodesk.Revit.DB.ElementId
    """

    ids = []
    ids_used = get_used_duct_and_flex_duct_symbol_ids(doc)
    ids_available = get_symbol_ids_for_duct_types_in_model(doc)
    for id in ids_available:
        if id not in ids_used:
            ids.append(id)
    return ids


def get_unused_duct_and_flex_duct_symbol_ids_for_purge(doc):
    """
    Gets all unused duct and flex duct symbol ids of categories
    - BuiltInCategory.OST_DuctAccessory,
    - BuiltInCategory.OST_DuctTerminal,
    - BuiltInCategory.OST_DuctFitting
    Unused: not one instance per symbol is placed in the model.
    This method can be used to safely delete unused duct symbols and families from the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of ids representing family symbols.
    :rtype: list  Autodesk.Revit.DB.ElementId
    """

    ids = rFamPurge.get_unused_in_place_ids_for_purge(
        doc, get_unused_duct_and_flex_duct_symbol_ids
    )
    return ids


def get_used_cable_tray_symbol_ids(doc):
    """
    Gets all used cable tray symbol ids of categories
    - BuiltInCategory.OST_CableTrayFitting
    Used: at least instance per symbol is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of ids representing family symbols.
    :rtype: list  Autodesk.Revit.DB.ElementId
    """

    ids = []
    ids_in_model = rPurgeUtils.get_used_unused_type_ids(
        doc, get_symbol_ids_for_cable_tray_types_in_model, 1
    )
    ids_used_in_types = get_symbol_ids_used_in_cable_tray_types(doc)
    ids = merge_into_unique_list(ids, ids_in_model)
    ids = merge_into_unique_list(ids, ids_used_in_types)
    return ids


def get_unused_cable_tray_symbol_ids(doc):
    """
    Gets all unused cable tray symbol ids of categories
    - BuiltInCategory.OST_CableTrayFitting
    Unused: not one instance per symbol is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of ids representing family symbols.
    :rtype: list  Autodesk.Revit.DB.ElementId
    """

    ids = []
    ids_used = get_used_cable_tray_symbol_ids(doc)
    ids_available = get_symbol_ids_for_cable_tray_types_in_model(doc)
    for id in ids_available:
        if id not in ids_used:
            ids.append(id)
    return ids


def get_unused_cable_tray_symbol_ids_for_purge(doc):
    """
    Gets all unused cable tray symbol ids of categories
    - BuiltInCategory.OST_CableTrayFitting
    Unused: not one instance per symbol is placed in the model.
    This method can be used to safely delete unused cable tray symbols and families from the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of ids representing family symbols.
    :rtype: list  Autodesk.Revit.DB.ElementId
    """

    ids = rFamPurge.get_unused_in_place_ids_for_purge(
        doc, get_unused_cable_tray_symbol_ids
    )
    return ids


def get_used_conduit_symbol_ids(doc):
    """
    Gets all used conduit symbol ids of categories
    - BuiltInCategory.OST_ConduitFitting
    Used: at least instance per symbol is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of ids representing family symbols.
    :rtype: list  Autodesk.Revit.DB.ElementId
    """

    ids = []
    ids_in_model = rPurgeUtils.get_used_unused_type_ids(
        doc, get_symbol_ids_for_conduit_types_in_model, 1
    )
    ids_used_in_types = get_symbol_ids_used_in_conduit_types(doc)
    ids = merge_into_unique_list(ids, ids_in_model)
    ids = merge_into_unique_list(ids, ids_used_in_types)
    return ids


def get_unused_conduit_symbol_ids(doc):
    """
    Gets all unused conduit symbol ids of categories
    - BuiltInCategory.OST_ConduitFitting
    Unused: not one instance per symbol is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of ids representing family symbols.
    :rtype: list  Autodesk.Revit.DB.ElementId
    """

    ids = []
    ids_used = get_used_conduit_symbol_ids(doc)
    ids_available = get_symbol_ids_for_conduit_types_in_model(doc)
    for id in ids_available:
        if id not in ids_used:
            ids.append(id)
    return ids


def get_unused_conduit_symbol_ids_for_purge(doc):
    """
    Gets all unused conduit symbol ids of categories
    - BuiltInCategory.OST_ConduitFitting
    Unused: not one instance per symbol is placed in the model.
    This method can be used to safely delete unused conduit symbols and families from the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of ids representing family symbols.
    :rtype: list  Autodesk.Revit.DB.ElementId
    """

    ids = rFamPurge.get_unused_in_place_ids_for_purge(
        doc, get_unused_conduit_symbol_ids
    )
    return ids


def get_used_pipe_symbol_ids(doc):
    """
    Gets all used pipe symbol ids of categories
    - BuiltInCategory.OST_PipeAccessory,
    - BuiltInCategory.OST_PipeFitting
    Used: at least instance per symbol is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of ids representing family symbols.
    :rtype: list  Autodesk.Revit.DB.ElementId
    """

    ids = []
    ids_in_model = rPurgeUtils.get_used_unused_type_ids(
        doc, get_symbol_ids_for_pipe_types_in_model, 1
    )
    ids_used_in_types = get_symbol_ids_used_in_pipe_types(doc)
    ids = merge_into_unique_list(ids, ids_in_model)
    ids = merge_into_unique_list(ids, ids_used_in_types)
    return ids


def get_unused_pipe_symbol_ids(doc):
    """
    Gets all unused pipe symbol ids of categories
    - BuiltInCategory.OST_PipeAccessory,
    - BuiltInCategory.OST_PipeFitting
    Unused: not one instance per symbol is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of ids representing family symbols.
    :rtype: list  Autodesk.Revit.DB.ElementId
    """

    ids = []
    ids_used = get_used_pipe_symbol_ids(doc)
    ids_available = get_symbol_ids_for_pipe_types_in_model(doc)
    for id in ids_available:
        if id not in ids_used:
            ids.append(id)
    return ids


def get_unused_pipe_symbol_ids_for_purge(doc):
    """
    Gets all unused pipe symbol ids of categories
    - BuiltInCategory.OST_PipeAccessory,
    - BuiltInCategory.OST_PipeFitting
    Unused: not one instance per symbol is placed in the model.
    This method can be used to safely delete unused pipe symbols and families from the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of ids representing family symbols.
    :rtype: list  Autodesk.Revit.DB.ElementId
    """

    ids = rFamPurge.get_unused_in_place_ids_for_purge(doc, get_unused_pipe_symbol_ids)
    return ids



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A class to store file select UI settings.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities.Objects import base

# An item to represent a file name in a row in a grid.
class FileSelectionSettings(base.Base):
    def __init__(
        self,
        input_path,
        include_sub_dirs_in_search,
        output_directory,
        output_file_number,
        revit_file_extension,
        filters,
        filter_type,
    ):
        """
        Class constructor

        :param inputDirectory: A fully qualified directory path containing files to be shown in UI.
        :type inputDirectory: str
        :param include_sub_dirs_in_search: If True include subdirectories in file search, otherwise just root directory.
        :type include_sub_dirs_in_search: bool
        :param output_directory: A fully qualified directory path to where task files will be written.
        :type output_directory: str
        :param output_file_number: The number of task files to be written out.
        :type output_file_number: int
        :param revitFileExtension: A file extension filter applied to directory search.
        :type revitFileExtension: str
        """

        self.input_directory = input_path
        self.incl_sub_dirs = include_sub_dirs_in_search
        self.output_dir = output_directory
        self.output_file_num = output_file_number
        self.revit_file_extension = revit_file_extension
        self.filters = filters
        self.filter_type = filter_type

        # ini super class to allow multi inheritance in children!
        super(FileSelectionSettings, self).__init__()



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a Revit ceilings geometry extraction functions. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#
import Autodesk.Revit.DB as rdb

from duHast.Revit.Ceilings import ceilings as rCeiling
from duHast.Data.Objects.Properties.Geometry import from_revit_conversion as rCon


def get_2d_points_from_revit_ceiling(ceiling):
    """
    Returns a list of lists of points representing the flattened(2D geometry) of the ceiling
    List of Lists because a ceiling can be made up of multiple sketches. Each nested list represents one ceiling sketch.
    Does not work with in place ceilings.

    :param ceiling: A revit ceiling instance.
    :type ceiling: Autodesk.Revit.DB.Ceiling

    :return: A list of data geometry instances.
    :rtype: list of :class:`.DataGeometry`
    """

    all_ceiling_points = []
    # get geometry from ceiling
    opt = rdb.Options()
    fr1_geom = ceiling.get_Geometry(opt)
    solids = []
    # check geometry for Solid elements
    # todo check for FamilyInstance geometry ( in place families!)
    for item in fr1_geom:
        if type(item) is rdb.Solid:
            solids.append(item)

    # process solids to points
    # in place families may have more then one solid
    for s in solids:
        point_per_ceilings = rCon.convert_solid_to_flattened_2d_points(s)
        if len(point_per_ceilings) > 0:
            for p_lists in point_per_ceilings:
                all_ceiling_points.append(p_lists)
    return all_ceiling_points


def get_2d_points_from_revit_ceilings_in_model(doc):
    """
    Returns a list of lists of points representing the flattened(2D geometry) of the ceiling
    List of Lists because a ceiling can be made up of multiple sketches. Each nested list represents one ceiling sketch.
    Does not work with in place ceilings.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A list of data geometry instances.
    :rtype: list of :class:`.DataGeometry`
    """

    ceiling_instances = rCeiling.get_all_ceiling_instances_in_model_by_category(doc)
    all_ceiling_points = []
    for c_i in ceiling_instances:
        ceiling_points = get_2d_points_from_revit_ceiling(c_i)
        if len(ceiling_points) > 0:
            all_ceiling_points.append(ceiling_points)
    return all_ceiling_points



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to changing Revit shared parameters to family parameters and vise versa.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


import Autodesk.Revit.DB as rdb

from duHast.Revit.Common import transaction as rTran
from duHast.Utilities.Objects import result as res
from duHast.Revit.SharedParameters.shared_parameters import get_family_parameters


def change_shared_parameter_to_family_parameter(doc, parameter_name, prefix="_"):
    """
    Changes a shared family parameter to a standard family parameter.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param parameter_name: The shared parameter name.
    :type parameter_name: str
    :param prefix: Revit requires the new parameter to have a different name to the shard parameter, therefore a prefix to the name is applied, defaults to '_'
    :type prefix: str, optional
    :return:
        Result class instance.
        - Parameter change status returned in result.status. False if an exception occurred, otherwise True.
        - result.message will contain the name of the shared parameter and the new family parameter name.
        - result.status will contain the new family parameter.
        On exception (handled by optimizer itself!):
        - result.status (bool) will be False.
        - result.message will contain generic exception message.
        - result.status will be an empty list
    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    # get the family manager
    manager = doc.FamilyManager
    # get family parameters
    paras = get_family_parameters(doc)
    # flag
    changed_parameter = False
    # check whether any parameter in family requires changing
    for p in paras:
        if p.Definition.Name == parameter_name:
            para_old_name = p.Definition.Name

            def action():
                action_return_value = res.Result()
                try:

                    parameter_new = manager.ReplaceParameter(
                        p,
                        prefix + para_old_name,
                        p.Definition.ParameterGroup,
                        p.IsInstance,
                    )

                    action_return_value.update_sep(
                        True,
                        para_old_name
                        + ": Successfully changed shared parameter to family parameter: "
                        + prefix
                        + para_old_name,
                    )
                    action_return_value.result.append(parameter_new)
                except Exception as e:
                    action_return_value.update_sep(
                        False,
                        para_old_name
                        + ": Failed to change shared parameter to family parameter: "
                        + str(e),
                    )
                return action_return_value

            transaction = rdb.Transaction(doc, "change to family parameter")
            return_value = rTran.in_transaction(transaction, action)
            changed_parameter = return_value.status
    if changed_parameter == False:
        return_value.status = False
        return_value.message = (
            "No parameter matching: "
            + parameter_name
            + " was found. No shared parameter was changed."
        )
    return return_value


def change_family_parameter_to_shared_parameter(
    doc, parameter_name, parameter_data, parameter_def
):
    """
    Changes a family parameter to a shared parameter.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param parameter_name: The family parameter name.
    :type parameter_name: str
    :param parameter_data: A named tup[le containing the shared parameter information
    :type parameter_data: RevitSharedParametersTuple.parameterData
    :param parameter_def: The external definition of the shared parameter.
    :type parameter_def: Autodesk.Revit.DB.ExternalDefinition
    :return:
        Result class instance.
        - Parameter change status returned in result.status. False if an exception occurred, otherwise True.
        - result.message will contain the name of the family parameter and the new shared parameter name.
        - result.status will contain the new shared parameter.
        On exception (handled by optimizer itself!):
        - result.status (bool) will be False.
        - result.message will contain generic exception message.
        - result.status will be an empty list
    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    # get the family manager
    manager = doc.FamilyManager
    # get family parameters
    paras = get_family_parameters(doc)
    # flag
    changed_parameter = False
    # check whether any parameter in family requires changing
    for p in paras:
        if p.Definition.Name == parameter_name:

            def action():
                action_return_value = res.Result()
                try:

                    parameter_new = manager.ReplaceParameter(
                        p,
                        parameter_def,
                        parameter_data.builtInParameterGroup,
                        parameter_data.isInstance,
                    )

                    action_return_value.update_sep(
                        True,
                        parameter_name
                        + ": Changed family parameter to shared parameter: "
                        + parameter_data.name,
                    )
                    action_return_value.result.append(parameter_new)
                except Exception as e:
                    action_return_value.update_sep(
                        False,
                        parameter_name
                        + ": Failed to change family parameter to shared parameter.",
                    )
                return action_return_value

            transaction = rdb.Transaction(doc, "change to shared parameter")
            return_value = rTran.in_transaction(transaction, action)
            changed_parameter = return_value.status
    if changed_parameter == False:
        return_value.status = False
        return_value.message = (
            "No parameter matching: "
            + parameter_name
            + " was found. No family parameter was changed."
        )
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Model health report functions - general reports.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Model health report metrics can either be displayed in a family where each parameter is assigned to a metric 
and or data can be exported to text files which can be used to visualize key metrics over time.

"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr


import System


clr.AddReference("System.Core")
from System import Linq

clr.ImportExtensions(Linq)

from duHast.Revit.ModelHealth.Reporting.Properties.constants import (
    FAILED_TO_RETRIEVE_VALUE,
)

from duHast.Utilities.date_stamps import get_file_date_stamp, FILE_DATE_STAMP_YYYY_MM_DD
from duHast.Utilities.files_io import file_exist, get_file_size
from duHast.Revit.BIM360.bim_360 import get_model_file_size
from duHast.Revit.Warnings.warnings import get_warnings
from duHast.Revit.Common.worksets import get_worksets

# --------------------------------------------- GENERAL ---------------------------------------------


def get_current_date(doc):
    """
    Get the current date

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: The current date in format YYYY_MM_DD.
    :rtype: str

    """
    return get_file_date_stamp(FILE_DATE_STAMP_YYYY_MM_DD)


def get_workset_number(doc):
    """
    Gets the number of worksets in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: The number of worksets in a model.
    :rtype: int
    """

    return len(get_worksets(doc))


def get_current_file_size(doc):
    """
    Gets the file size in MB.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: File size in MB. On exception it will return -1
    :rtype: int
    """

    size = FAILED_TO_RETRIEVE_VALUE
    try:
        # get the path from the document
        # this will fail if not a file based doc or the document is detached
        revit_file_path = doc.PathName
        try:
            # test if this is a cloud model
            path = doc.GetCloudModelPath()
            size = get_model_file_size(doc)
        except:
            # local file server model
            if file_exist(revit_file_path):
                # get file size in MB
                size = get_file_size(revit_file_path)
            else:
                raise ValueError("File not found: {}".format(revit_file_path))
    except Exception as e:
        raise ValueError("Failed to get file size with: {}".format(e))
    return size


def get_number_of_warnings(doc):
    """
    Gets the number of warnings in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: Number of warnings in model. On exception it will return -1
    :rtype: int
    """

    number = FAILED_TO_RETRIEVE_VALUE
    try:
        number = len(get_warnings(doc))
    except Exception as e:
        raise ValueError("Failed to get number of warnings: {}".format(e))
    return number



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of functions around Revit Design Sets and Design Options.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr
import System

# import common library modules
#from duHast.Revit.Common import parameter_get_utils as rParaGet


# import Autodesk
from Autodesk.Revit.DB import BuiltInParameter, DesignOption, Element, FilteredElementCollector

# -------------------------------------------- common variables --------------------
#: header used in reports
REPORT_DESIGNSET_HEADER = ["HOSTFILE", "ID", "NAME", "PRIMARY OPTION", "OTHER OPTIONS"]

# --------------------------------------------- utility functions ------------------


def get_design_options(doc):
    """
    Gets all design options in a model,

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: Design options in current model
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = FilteredElementCollector(doc).OfClass(DesignOption)
    return collector


def get_design_sets(doc):
    """
    Gets all the design sets in a model,

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: Design sets in the current model
    :rtype: list of Autodesk.Revit.DB.Element
    """

    collector = get_design_options(doc=doc)
    design_sets = []
    design_set_names = []
    for do in collector:
        e = doc.GetElement(
            do.get_Parameter(BuiltInParameter.OPTION_SET_ID).AsElementId()
        )
        designSetName = Element.Name.GetValue(e)
        if designSetName not in design_set_names:
            design_sets.append(e)
            design_set_names.append(designSetName)
    return design_sets


def is_design_option_primary(doc, design_set_name, design_option_name):
    """
    Checks whether a design option is the primary option within a design set.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param design_set_name: The name of the design set the option belongs to,
    :type design_set_name: str
    :param design_option_name: The name of the design option to be checked,
    :type design_option_name: str
    :return: True if this option is primary otherwise False
    :rtype: bool
    """

    collector = get_design_options(doc=doc)
    is_primary = False
    # loop over all design options in model, get the set they belong to and check for matches on both, set and option, by name
    for do in collector:
        design_o_name = Element.Name.GetValue(do)
        # check if '< 'in name indicating a primary option, if so remove from name
        index_chevron = design_o_name.find("<")
        if index_chevron > 0:
            design_o_name = design_o_name[: index_chevron - 2]
        # design set
        design_set = doc.GetElement(
            do.get_Parameter(BuiltInParameter.OPTION_SET_ID).AsElementId()
        )
        design_set_name = Element.Name.GetValue(design_set)
        # check for match on both set and option
        if design_set_name == design_set_name and design_o_name == design_option_name:
            # get isPrimary property on design option
            is_primary = do.IsPrimary
            break
    return is_primary


def get_design_set_option_info(doc, element):
    """
    Get the design set, design option information of an element.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param element: The element of which the design set/option data is to be returned.
    :type element: Autodesk.Revit.DB.Element
    :return: Dictionary
        Design Set Name: (can be either Main Model or the design set name)
        designOptionName:    Design Option Name (empty string if Main Model
        isPrimary:           Indicating whether design option is primary (true also if Main Model)
    :rtype: Dictionary
        designSetName:str
        designOptionName:str
        isPrimary:bool
    """

    # keys match properties in DataDesignSetOption class!!
    new_key = ["designSetName", "designOptionName", "isPrimary"]
    new_value = ["Main Model", "-", True]
    dic = dict(zip(new_key, new_value))
    try:
        # this only works for objects inheriting from Autodesk.Revit.DB.Element
        design_option = element.DesignOption
        dic["designOptionName"] = design_option.Name
        dic["isPrimary"] = design_option.IsPrimary
        e = doc.GetElement(
            design_option.get_Parameter(
                BuiltInParameter.OPTION_SET_ID
            ).AsElementId()
        )
        dic["designSetName"] = Element.Name.GetValue(e)
    except Exception as e:
        pass
    return dic



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Helper functions relating to combining text files. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


import codecs
import glob
import os
from duHast.Utilities.files_io import get_file_name_without_ext
from duHast.Utilities.files_get import get_files_single_directory
from duHast.Utilities.files_tab import get_unique_headers as get_unique_headers_tab
from duHast.Utilities.files_tab import read_tab_separated_file, write_report_data
from duHast.Utilities.files_csv import get_unique_headers as get_unique_headers_csv
from duHast.Utilities.files_csv import read_csv_file, write_report_data_as_csv
from duHast.Utilities.files_json import read_json_data_from_file, write_json_to_file


def combine_files(
    folder_path,
    file_prefix="",
    file_suffix="",
    file_extension=".txt",
    out_put_file_name="result.txt",
    file_getter=get_files_single_directory,
):
    """
    Combines multiple text files into a single new file.
    Assumes:

    - files have a header row followed by data rows
    - same number of headers (columns) in each files.
    - files have the same header names per column

    The new file will be saved into the same folder as the original files.

    :param folder_path: Folder path from which to get files to be combined and to which the combined file will be saved.
    :type folder_path: str
    :param file_prefix: Filter: File name starts with this value
    :type file_prefix: str
    :param file_suffix: Filter: File name ends with this value.
    :type file_suffix: str
    :param file_extension: Filter: File needs to have this file extension
    :type file_extension: str, format '.extension'
    :param out_put_file_name: The file name of the combined file, defaults to 'result.txt'
    :type out_put_file_name: str, optional
    :param file_getter: Function returning list of files to be combined, defaults to GetFilesSingleFolder
    :type file_getter: func(folder_path, file_prefix, file_suffix, file_extension), optional
    """

    file_list = file_getter(folder_path, file_prefix, file_suffix, file_extension)
    with open(os.path.join(folder_path, out_put_file_name), "w") as result:
        file_counter = 0
        for file_ in file_list:
            line_counter = 0
            fp = open(file_, "r")
            lines = fp.readlines()
            fp.close()
            for line in lines:
                # ensure header from first file is copied over
                if file_counter == 0 and line_counter == 0 or line_counter != 0:
                    result.write(line)
                line_counter += 1

            file_counter += 1


def combine_files_basic(
    folder_path,
    file_prefix="",
    file_suffix="",
    file_extension=".txt",
    out_put_file_name="result.txt",
    file_getter=get_files_single_directory,
):
    """
    Combines multiple text files into a single new file.
    Assumes:

    - files are text files

    The new file will be saved into the same folder as the original files.

    :param folder_path: Folder path from which to get files to be combined and to which the combined file will be saved.
    :type folder_path: str
    :param file_prefix: Filter: File name starts with this value
    :type file_prefix: str
    :param file_suffix: Filter: File name ends with this value.
    :type file_suffix: str
    :param file_extension: Filter: File needs to have this file extension
    :type file_extension: str, format '.extension'
    :param out_put_file_name: The file name of the combined file, defaults to 'result.txt'
    :type out_put_file_name: str, optional
    :param file_getter: Function returning list of files to be combined, defaults to GetFilesSingleFolder
    :type file_getter: func(folder_path, file_prefix, file_suffix, file_extension), optional
    """

    file_list = file_getter(folder_path, file_prefix, file_suffix, file_extension)
    with open(os.path.join(folder_path, out_put_file_name), "w") as f:
        for file_ in file_list:
            fp = open(file_, "r")
            lines = fp.readlines()
            fp.close()
            for line in lines:
                f.write(line)
        f.close()


def append_to_file(source_file, append_file, ignore_first_row=False):
    """
    Appends one text file to another. Assumes same number of headers (columns) in both files.

    :param source_file: The fully qualified file path of the file to which the other file will be appended.
    :type source_file: str
    :param append_file: The fully qualified file path of the file to be appended.
    :type append_file: str
    :param ignore_first_row: If True, first row of append file will not be appended to source file.
    :type ignore_first_row: bool
    :return: If True file was appended without an exception, otherwise False.
    :rtype: bool
    """

    flag = True
    try:
        # read file to append into memory...hopefully will never get in GB range in terms of file size
        fp = codecs.open(append_file, "r", encoding="utf-8")
        lines = fp.readlines()
        fp.close()
        with codecs.open(source_file, "a", encoding="utf-8") as f:
            if ignore_first_row == False:
                for line in lines:
                    f.write(line)
            else:
                # check if only a header row in file?
                if len(lines) > 1:
                    for x in range(1, len(lines)):
                        f.write(lines[x])
                        # get out to avoid out of index exception
                        # if I change the for loop to be
                        # for x in range(1, len(lines)-1):
                        # it will not process files with two rows only ( a header and a single data row! )
                        if x == len(lines) - 1:
                            break

    except Exception:
        flag = False
    return flag


def _format_headers(headers_in_file, file):
    """
    Replace any empty strings in header row

    :param headers_in_file: list of header entries
    :type headers_in_file: [str]
    :param file: Fully qualified file name
    :type file: str

    :return: Header row
    :rtype:[str]
    """

    file_name = get_file_name_without_ext(file)
    empty_header_counter = 0
    for i in range(len(headers_in_file)):
        # reformat any empty headers to be unique
        if headers_in_file[i] == "":
            headers_in_file[i] = file_name + ".Empty." + str(empty_header_counter)
            empty_header_counter = empty_header_counter + 1
    return headers_in_file


def combine_files_header_independent(
    folder_path,
    file_prefix="",
    file_suffix="",
    file_extension=".txt",
    out_put_file_name="result.txt",
):
    """
    Used to combine report files into one file, files may have different number / named columns.

    Columns which are unique to some files will have as a value 'N/A' in files where those columns do not exist.
    File need to use <tab> character as column separator

    :param folder_path: Folder path from which to get files to be combined and to which the combined file will be saved.
    :type folder_path: str
    :param file_prefix: Filter: File name starts with this value
    :type file_prefix: str
    :param file_suffix: Filter: File name ends with this value.
    :type file_suffix: str
    :param file_extension: Filter: File needs to have this file extension
    :type file_extension: str, format '.extension'
    :param out_put_file_name: The file name of the combined file, defaults to 'result.txt'
    :type out_put_file_name: str, optional
    """

    file_list = glob.glob(
        folder_path + "\\" + file_prefix + "*" + file_suffix + file_extension
    )
    # build list of unique headers
    headers = get_unique_headers_tab(file_list)
    combined_file_name = os.path.join(folder_path, out_put_file_name)
    # loop over files to be combined
    file_counter = 0
    for file in file_list:
        line_counter = 0
        column_mapper = []
        lines = read_tab_separated_file(file)
        lines_to_be_transferred = []
        for line in lines:
            # read the headers in file
            if line_counter == 0:
                # replace any empty strings in header
                headers_in_file = _format_headers(line, file)
                # match up unique headers with headers from this file
                # build header mapping
                for unique_header in headers:
                    if unique_header in headers_in_file:
                        column_mapper.append(headers_in_file.index(unique_header))
                    else:
                        column_mapper.append(-1)
            # ensure unique header is written to file
            if file_counter == 0 and line_counter == 0:
                lines_to_be_transferred.append(headers)
            elif line_counter != 0:
                padded_row = []
                for cm in column_mapper:
                    if cm == -1:
                        # this column does not exist in this file
                        padded_row.append("N/A")
                    elif cm > len(line):
                        # less columns in file than mapper index (should'nt happen??)
                        padded_row.append("index out of bounds")
                    else:
                        padded_row.append(line[cm])
                lines_to_be_transferred.append(padded_row)
            line_counter += 1
        # write file data to combined file
        write_report_data(
            combined_file_name, header=[], data=lines_to_be_transferred, write_type="a"
        )
        file_counter += 1


def combine_files_csv_header_independent(
    folder_path,
    file_prefix="",
    file_suffix="",
    file_extension=".txt",
    out_put_file_name="result.csv",
):
    """
    Used to combine report files into one file, files may have different number / named columns.

    Columns which are unique to some files will have as a value 'N/A' in files where those columns do not exist.
    File need to use <,> character as column separator. (.CSV)

    :param folder_path: Folder path from which to get files to be combined and to which the combined file will be saved.
    :type folder_path: str
    :param file_prefix: Filter: File name starts with this value
    :type file_prefix: str
    :param file_suffix: Filter: File name ends with this value.
    :type file_suffix: str
    :param file_extension: Filter: File needs to have this file extension
    :type file_extension: str, format '.extension'
    :param out_put_file_name: The file name of the combined file, defaults to 'result.csv'
    :type out_put_file_name: str, optional
    """

    file_list = glob.glob(
        folder_path + "\\" + file_prefix + "*" + file_suffix + file_extension
    )
    # build list of unique headers
    headers = get_unique_headers_csv(file_list)
    combined_file_name = os.path.join(folder_path, out_put_file_name)

    # loop over files and combine...
    file_counter = 0
    for file in file_list:
        line_counter = 0
        column_mapper = []
        lines = read_csv_file(file, increaseMaxFieldSizeLimit=False)
        lines_to_be_transferred = []
        for line in lines:
            # read the headers in file
            if line_counter == 0:
                # replace any empty strings in header
                headers_in_file = _format_headers(line, file)
                # match up unique headers with headers from this file
                # build header mapping
                for unique_header in headers:
                    if unique_header in headers_in_file:
                        column_mapper.append(headers_in_file.index(unique_header))
                    else:
                        column_mapper.append(-1)
            # ensure unique header is written to file
            if file_counter == 0 and line_counter == 0:
                lines_to_be_transferred.append(headers)
            elif line_counter != 0:
                # map data columns to headers
                padded_row = []
                for cm in column_mapper:
                    if cm == -1:
                        # this column does not exist in this file
                        padded_row.append("N/A")
                    elif cm > len(line):
                        # less columns in file than mapper index (should'nt happen??)
                        padded_row.append("index out of bounds")
                    else:
                        padded_row.append(line[cm])
                lines_to_be_transferred.append(padded_row)
            line_counter += 1
        # write file data to combined file
        write_report_data_as_csv(
            combined_file_name, header=[], data=lines_to_be_transferred, write_type="a"
        )
        file_counter += 1


def combine_files_json(
    folder_path,
    file_prefix="",
    file_suffix="",
    file_extension=".txt",
    out_put_file_name="result.txt",
    file_getter=get_files_single_directory,
):
    """
    Combines multiple json formatted text files into a single json list formatted file, where each file is a list entry.
    Assumes:

    - each file can contain a single line json formatted string

    The new file will be saved into the same folder as the original files.

    :param folder_path: Folder path from which to get files to be combined and to which the combined file will be saved.
    :type folder_path: str
    :param file_prefix: Filter: File name starts with this value
    :type file_prefix: str
    :param file_suffix: Filter: File name ends with this value.
    :type file_suffix: str
    :param file_extension: Filter: File needs to have this file extension
    :type file_extension: str, format '.extension'
    :param out_put_file_name: The file name of the combined file, defaults to 'result.txt'
    :type out_put_file_name: str, optional
    :param file_getter: Function returning list of files to be combined, defaults to GetFilesSingleFolder
    :type file_getter: func(folder_path, file_prefix, file_suffix, file_extension), optional
    """

    # get all files to be combined
    file_list = file_getter(folder_path, file_prefix, file_suffix, file_extension)

    # read json data into a list of json objects
    json_objects = []
    for file in file_list:
        json_object = read_json_data_from_file(file_path=file)
        json_objects.append(json_object)

    # write json data out
    result_write = write_json_to_file(
        json_data=json_objects,
        data_output_file_path=os.path.join(folder_path, out_put_file_name),
    )

    # return flag only
    return result_write.status


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Helper functions relating to tab separated text files. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import codecs
import csv
from duHast.Utilities.files_io import get_first_row_in_file_no_strip
from duHast.Utilities.files_io import get_file_name_without_ext


def get_unique_headers(files):
    """
    Gets a list of alphabetically sorted headers retrieved from text files.
    Assumes:

    - first row in each file is the header row
    - headers are separated by <tab> character

    :param files: List of file path from which the headers are to be returned.
    :type files: list of str
    :return: List of headers.
    :rtype: list of str
    """

    headers_in_all_files = {}
    for f in files:
        # get unmodified row data and remove the next line character at the end
        data = get_first_row_in_file_no_strip(f).rstrip("\n")
        if data is not None:
            row_split = data.split("\t")
            headers_in_all_files[get_file_name_without_ext(f)] = row_split
    headers_unique = []
    for header_by_file in headers_in_all_files:
        empty_header_counter = 0
        for header in headers_in_all_files[header_by_file]:
            # reformat any empty headers to be unique
            if header == "":
                header = header_by_file + ".Empty." + str(empty_header_counter)
                empty_header_counter = empty_header_counter + 1
            if header not in headers_unique:
                headers_unique.append(header)
    return sorted(headers_unique)


def write_report_data(file_name, header, data, write_type="w"):
    """
    Function writing out report information.

    :param file_name: The reports fully qualified file path.
    :type file_name: str
    :param header: list of column headers
    :type header: list of str
    :param data: list of list of strings representing row data
    :type data: [[str,str,..]]
    :param write_type: Flag indicating whether existing report file is to be overwritten 'w' or appended to 'a', defaults to 'w'
    :type write_type: str, optional
    """

    with codecs.open(file_name, write_type, encoding="utf-8") as f:
        # check if header is required
        if len(header) > 0:
            f.write("\t".join(header + ["\n"]))
        # check if data is required
        if len(data) > 0:
            for d in data:
                if len(d) > 1:
                    f.write("\t".join(d) + "\n")
                elif len(d) == 1:
                    f.write(d[0] + "\n")
        f.close()


def read_tab_separated_file(file_path, increase_max_field_size_limit=False):
    """
    Read a tab separated text file into a list of rows, where each row is another list.
    :param file_path: The fully qualified file path to the tab separated text file.
    :type file_path: str
    :return:  A list of list of strings representing the data in each row.
    :rtype: list of list of str
    """

    row_list = []

    # hard coded hack
    if increase_max_field_size_limit:
        csv.field_size_limit(2147483647)

    try:
        with open(file_path) as f:
            reader = csv.reader(f, dialect="excel-tab")
            for row in reader:  # each row is a list
                row_list.append(row)
            f.close()
    except Exception as e:
        print(file_path, str(e))
        row_list = []
    return row_list



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

# import clr
# import System
# from numpy import empty
import os
import os.path
import collections

# import clr
# clr.AddReference("System.Core")
# from System import Linq
# clr.ImportExtensions(Linq)


def get_local_app_data_path():
    """
    return directory path to local app data folder

    :return: Path to local app data
    :rtype: str
    """

    return os.environ["LOCALAPPDATA"]


def get_current_user_name():
    """
    Returns the current user name

    :return: the user name
    :rtype: str
    """

    return os.environ["USERNAME"]


# ---------------------------------------------------------------------------------------------------------------------------------


def parse_string_to_bool(text):
    """
    Converts a string to lower case and then to bool. Will throw an exception if it fails to do so.

    ( 'true' = True, 'false' = False)

    :param text: The string representing a bool.
    :type text: str
    :raises Exception: If string to bool conversion fails the 'String cant be converted to bool' exception will be raised.

    :return: True or False
    :rtype: bool
    """

    if text.lower() == "true":
        return True
    elif text.lower() == "false":
        return False
    else:
        raise Exception("String cant be converted to bool")


# ---------------------------------------------------------------------------------------------------------------------------------

#: two digit padding
PAD_SINGLE_DIGIT_TO_TWO = "%02d"
#: three digit padding
PAD_SINGLE_DIGIT_TO_THREE = "%03d"


def pad_single_digit_numeric_string(numericString, format=PAD_SINGLE_DIGIT_TO_TWO):
    """
    Pads a single digit integer (past in as a string) with a leading zero (default)

    :param numericString: Integer as string.
    :type numericString: str
    :param format: The integer padding format, defaults to PAD_SINGLE_DIGIT_TO_TWO
    :type format: str, optional

    :return: The padded integer as string.
    :rtype: str
    """

    # attempt to convert string to int first
    try:
        value = int(numericString)
        return str(format % value)
    except Exception:
        # string was not an integer...
        return numericString


def encode_ascii(string):
    """
    Encode a string as ascii and replaces all non ascii characters

    If a non string is past in the value will be returned unchanged.

    :param string: The string to be ascii encoded.
    :type string: str

    :return: ascii encoded string
    :rtype: str
    """
    if type(string) == str:
        return string.encode("ascii", "replace")
    else:
        return string


def get_first(iterable, default, condition=lambda x: True):
    """
    Returns the first value in a list matching condition. If no value found returns the specified default value.

    :param iterable: the list to be searched.
    :type iterable: iterable
    :param default: The default value
    :type default: var
    :param condition: The condition to be checked, defaults to lambda x:True
    :type condition: _type_, optional

    :return: First value matching condition, otherwise default value
    :rtype: var
    """

    return next((x for x in iterable if condition(x)), default)


def index_of(list, item):
    """
    Gets the index of item in list

    :param list: The list
    :type list: list
    :param item: The item of which to return the index.
    :type item: var

    :return: The index of the item in the list, if no match -1 will be returned
    :rtype: int
    """
    try:
        return list.index(item)
    except:
        return -1


def remove_items_from_list(sourceList, removeIdsList):
    """
    helper removes items from a source list

    :param sourceList: The list containing items
    :type sourceList: list var
    :param removeIdsList: the list containing items to be removed
    :type removeIdsList: list var

    :return: The filtered source list.
    :rtype: list var
    """

    try:
        for item in removeIdsList:
            sourceList.remove(item)
    except:
        pass
    return sourceList


def flatten(d, parent_key="", sep="_"):
    """
    Flattens a dictionary as per stack overflow

    https://stackoverflow.com/questions/6027558/flatten-nested-dictionaries-compressing-keys/6027615#6027615

    :param d: _description_
    :type d: _type_
    :param parent_key: _description_, defaults to ''
    :type parent_key: str, optional
    :param sep: _description_, defaults to '_'
    :type sep: str, optional
    :return: _description_
    :rtype: _type_
    """

    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, collections.MutableMapping):
            items.extend(flatten(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)

def flatten_list(lst):
    """
    Function to flatten a nested list
    
    :param lst: The list to be flattened
    :type lst: list
    :return: The flattened list
    :rtype: list
    """
    flattened = []
    for item in lst:
        if isinstance(item, list):
            flattened.extend(flatten_list(item))
        else:
            flattened.append(item)
    return flattened


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains the Revit category report files IO functionality.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from duHast.Revit.Categories.Objects.object_style import ObjectStyle
from duHast.Utilities.files_json import write_json_to_file, read_json_data_from_file
from duHast.Revit.Categories.Reporting.categories_styles_model_json_props import (
    PROP_FILE_NAME,
    PROP_CATEGORY_STYLE_DATA,
)


def write_category_graphics_settings_report(revit_file_name, file_path, data):
    """
    Write category graphic settings to file.

    :param revit_file_name: The revit file name (without path or file extension)
    :type revit_file_name: str
    :param file_path: Fully qualified file path
    :type file_path: str
    :param data: json dictionary to be written to file
    :type data: {str:Any}

    :return:
        Result class instance.

        - result.status False if file failed to write, otherwise True.
        - result.message will contain file name of file written.
        - result.result: empty list

        On exception:

        - result.status (bool) will be False.
        - result.message will contain the exception message.
        - result.result: will be an empty list

    :rtype: :class:`.Result`
    """

    # wrap data to include file name
    json_data = {PROP_FILE_NAME: revit_file_name, PROP_CATEGORY_STYLE_DATA: data}

    result = write_json_to_file(json_data=json_data, data_output_file_path=file_path)
    return result

def read_category_graphics_data_from_file(file_path):
    """
    Reads a category graphics data report file into a list of ObjectStyle instances

    :param file_path: Fully qualified file path of report file.
    :type file_path: str
    :raises ValueError: If data node is missing from file

    :return: list of settings if files was read successfully otherwise an exception will be raised.
    :rtype: [:class:`.ObjectStyle`]
    """

    data_views = []
    # read json file
    data_read = read_json_data_from_file(file_path=file_path)

    # check if this is a list of data view items
    if isinstance(data_read, list):
        for data_entry in data_read:
            # check it got the required property node
            if (PROP_CATEGORY_STYLE_DATA) in data_entry:
                # convert node entries into class instances
                for entry in data_entry[PROP_CATEGORY_STYLE_DATA]:
                    data_view = ObjectStyle(j=entry)
                    data_views.append(data_view)
    elif isinstance(data_read, dict):
        # check it got the required property node
        if (PROP_CATEGORY_STYLE_DATA) in data_read:
            # convert node entries into class instances
            for entry in data_read[PROP_CATEGORY_STYLE_DATA]:
                data_view = ObjectStyle(j=entry)
                data_views.append(data_view)
    else:
        # missing node...raise an exception
        raise ValueError("Data does not contain a {} node".format(PROP_CATEGORY_STYLE_DATA))

    return data_views


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to purging Revit grids and grid heads. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import Autodesk.Revit.DB as rdb

from duHast.Revit.Family import purge_unused_family_types as rFamUPurge
from duHast.Revit.Common import (
    parameter_get_utils as rParaGet,
    purge_utils as rPurgeUtils,
)
from duHast.Revit.Grids import grids as rGrids


def get_unused_grid_types_for_purge(doc):
    """this will return all ids of unused grid types in the model to be purged"""
    return rPurgeUtils.get_used_unused_type_ids(
        doc, rGrids.get_all_grid_type_ids_by_category, 0, 8
    )


def get_unused_grid_head_families(doc):
    """this will return all ids of unused family symbols (types) of grid head families"""
    used_type_ids = rPurgeUtils.get_used_unused_type_ids(
        doc, rGrids.get_all_grid_type_ids_by_category, 1, 8
    )
    heads_in_use_ids = []
    for id in used_type_ids:
        type = doc.GetElement(id)
        id = rParaGet.get_built_in_parameter_value(
            type, rdb.BuiltInParameter.GRID_HEAD_TAG
        )
        if id != None and id not in heads_in_use_ids:
            heads_in_use_ids.append(id)
    all_symbols_in_model = rGrids.get_all_grid_heads_by_category(doc)
    unused_symbol_ids = []
    for symbol_in_model in all_symbols_in_model:
        if symbol_in_model.Id not in heads_in_use_ids:
            unused_symbol_ids.append(symbol_in_model.Id)
    return unused_symbol_ids


def get_unused_grid_head_families_for_purge(doc):
    """this will return all ids of unused grid head symbols and families to be purged"""
    return rFamUPurge.get_unused_in_place_ids_for_purge(
        doc, get_unused_grid_head_families
    )



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a Revit ceilings export to DATA class functions. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import Autodesk.Revit.DB as rdb

from duHast.Revit.Common import (
    design_set_options as rDesignO,
    parameter_get_utils as rParaGet,
    phases as rPhase,
)

# from duHast.APISamples.Ceilings.Geometry import Geometry
from duHast.Revit.Ceilings import ceilings as rCeiling
from duHast.Data.Objects import data_ceiling as dCeiling
from duHast.Data.Objects.Properties.Geometry import from_revit_conversion as rCon
from duHast.Revit.Common.Geometry import solids as rSolid


def populate_data_ceiling_object(doc, revit_ceiling):
    """
    Returns a custom ceiling data objects populated with some data from the revit model ceiling past in.

    - ceiling id
    - ceiling type name
    - ceiling mark
    - ceiling type mark
    - ceiling level name
    - ceiling level id
    - ceiling offset from level

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param revit_ceiling: A revit ceiling instance.
    :type revit_ceiling: Autodesk.Revit.DB.Ceiling

    :return: A data ceiling object instance.
    :rtype: :class:`.DataCeiling`
    """

    # set up data class object
    data_c = dCeiling.DataCeiling()
    # get ceiling geometry (boundary points)
    revit_geometry_point_groups = rSolid.get_2d_points_from_solid(revit_ceiling)
    # revitGeometryPointGroups = Geometry.Get2DPointsFromRevitCeiling(revitCeiling)
    if len(revit_geometry_point_groups) > 0:
        ceiling_point_groups_as_doubles = []
        for all_ceiling_point_groups in revit_geometry_point_groups:
            data_geo_converted = rCon.convert_xyz_in_data_geometry_polygons(
                doc, all_ceiling_point_groups
            )
            ceiling_point_groups_as_doubles.append(data_geo_converted)
        data_c.polygon = ceiling_point_groups_as_doubles
        # get design set data
        design_set_data = rDesignO.get_design_set_option_info(doc, revit_ceiling)
        data_c.design_set_and_option.option_name = design_set_data["designOptionName"]
        data_c.design_set_and_option.set_name = design_set_data["designSetName"]
        data_c.design_set_and_option.is_primary = design_set_data["isPrimary"]

        # get type properties
        data_c.type_properties.id = revit_ceiling.GetTypeId().IntegerValue
        data_c.type_properties.name = rdb.Element.Name.GetValue(revit_ceiling).encode(
            "utf-8"
        )
        ceiling_type = doc.GetElement(revit_ceiling.GetTypeId())

        # custom parameter value getters
        value_getter = {
            rdb.StorageType.Double: rParaGet.getter_double_as_double_converted_to_metric,
            rdb.StorageType.Integer: rParaGet.getter_int_as_int,
            rdb.StorageType.String: rParaGet.getter_string_as_UTF8_string,  # encode ass utf 8 just in case
            rdb.StorageType.ElementId: rParaGet.getter_element_id_as_element_int,  # needs to be an integer for JSON encoding
            str(None): rParaGet.getter_none,
        }
        data_c.type_properties.properties = (
            rParaGet.get_all_parameters_and_values_wit_custom_getters(
                ceiling_type, value_getter
            )
        )

        # get instance properties
        data_c.instance_properties.id = revit_ceiling.Id.IntegerValue
        data_c.instance_properties.properties = (
            rParaGet.get_all_parameters_and_values_wit_custom_getters(
                revit_ceiling, value_getter
            )
        )

        # get level properties
        data_c.level.name = rdb.Element.Name.GetValue(
            doc.GetElement(revit_ceiling.LevelId)
        ).encode("utf-8")
        data_c.level.id = revit_ceiling.LevelId.IntegerValue
        data_c.level.offset_from_level = rParaGet.get_built_in_parameter_value(
            revit_ceiling, rdb.BuiltInParameter.CEILING_HEIGHTABOVELEVEL_PARAM
        )  # offset from level

        # get the model name
        if doc.IsDetached:
            data_c.revit_model.name = "Detached Model"
        else:
            data_c.revit_model.name = doc.Title

        # get phasing information
        data_c.phasing.created = rPhase.get_phase_name_by_id(
            doc,
            rParaGet.get_built_in_parameter_value(
                revit_ceiling,
                rdb.BuiltInParameter.PHASE_CREATED,
                rParaGet.get_parameter_value_as_element_id,
            ),
        ).encode("utf-8")
        data_c.phasing.demolished = rPhase.get_phase_name_by_id(
            doc,
            rParaGet.get_built_in_parameter_value(
                revit_ceiling,
                rdb.BuiltInParameter.PHASE_DEMOLISHED,
                rParaGet.get_parameter_value_as_element_id,
            ),
        ).encode("utf-8")

        return data_c
    else:
        return None


def get_all_ceiling_data(doc):
    """
    Gets a list of ceiling data objects for each ceiling element in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A list of data ceiling instances.
    :rtype: list of :class:`.DataCeiling`
    """

    all_ceiling_data = []
    ceilings = rCeiling.get_all_ceiling_instances_in_model_by_category(doc)
    for ceiling in ceilings:
        cd = populate_data_ceiling_object(doc, ceiling)
        if cd is not None:
            all_ceiling_data.append(cd)
    return all_ceiling_data



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Family line pattern data processor class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from duHast.Revit.Family.Data.Objects.ifamily_processor import IFamilyProcessor
from duHast.Revit.LinePattern.Data.Objects import line_pattern_data as rLinePatData
from duHast.Revit.Family.Data.Objects import ifamily_data as IFamData
from duHast.Utilities.Objects import result as res


class LinePatternProcessor(IFamilyProcessor):
    def __init__(self, pre_actions=None, post_actions=None):
        """
        Class constructor.
        """

        # setup report header
        string_report_headers = [
            IFamData.ROOT,
            IFamData.ROOT_CATEGORY,
            IFamData.FAMILY_NAME,
            IFamData.FAMILY_FILE_PATH,
            IFamData.USAGE_COUNTER,
            IFamData.USED_BY,
            rLinePatData.PATTERN_NAME,
            rLinePatData.PATTERN_ID,
        ]

        # store data type  in base class
        super(LinePatternProcessor, self).__init__(
            pre_actions=pre_actions,
            post_actions=[self._post_action_update_used_line_patterns],
            data_type="LinePattern",
            string_report_headers=string_report_headers,
        )

        # set default post action to updated line patterns used in root processor with any line patterns found in nested
        # families
        # self.postActions = [self._postActionUpdateUsedLinePatterns]
        # add any other post actions
        if post_actions != None:
            for post_action in post_actions:
                self.post_actions.append(post_action)

    def process(self, doc, root_path, root_category_path):
        """
        Calls processor instance with the document and root path provided and adds processor instance to class property .data

        :param doc: Current family document.
        :type doc: Autodesk.Revit.DB.Document
        :param rootPath: The path of the nested family in a tree: rootFamilyName::nestedFamilyNameOne::nestedFamilyTwo\
            This includes the actual family name as the last node.
        :type rootPath: str
        :param rootCategoryPath: The categroy path of the nested family in a tree: rootFamilyCategory::nestedFamilyOneCategory::nestedFamilyTwoCategory\
            This includes the actual family category as the last node.
        :type rootCategoryPath: str
        """

        dummy = rLinePatData.LinePatternData(
            root_path, root_category_path, self.data_type
        )
        dummy.process(doc)
        self.data.append(dummy)

    def _is_sub_line_pattern_present(
        self, root_family_data, nested_family_line_pattern
    ):
        match = None
        for root_fam in root_family_data:
            if (
                root_fam[rLinePatData.PATTERN_NAME]
                == nested_family_line_pattern[rLinePatData.PATTERN_NAME]
            ):
                match = root_fam
                break
        return match

    def _update_root_family_data(self, root_family_data, nested_families_line_patterns):
        # loop over nested family line pattern data
        for nested_line_pattern in nested_families_line_patterns:
            # check if pattern is already in root family
            matching_root_fam_pattern = self._is_sub_line_pattern_present(
                root_family_data, nested_line_pattern
            )
            if matching_root_fam_pattern != None:
                # update used by list
                if (
                    nested_line_pattern[rLinePatData.PATTERN_NAME]
                    not in matching_root_fam_pattern[IFamData.USED_BY]
                ):
                    # add the root path to the used by list for ease of identification of the origin of this pattern usage
                    matching_root_fam_pattern[IFamData.USED_BY].append(
                        {
                            rLinePatData.PATTERN_ID: nested_line_pattern[
                                rLinePatData.PATTERN_ID
                            ],
                            IFamData.ROOT: nested_line_pattern[IFamData.ROOT],
                        }
                    )
                    # update used by counter
                    matching_root_fam_pattern[IFamData.USAGE_COUNTER] = (
                        matching_root_fam_pattern[IFamData.USAGE_COUNTER] + 1
                    )
            else:
                pass
                # nothing to do if that pattern has not been reported to start off with
                # this patter could, for example, belong to the section marker family present in most 3d families

    def _get_used_line_patterns(self, data):
        used_line_patterns = []
        for d in data:
            if d[IFamData.USAGE_COUNTER] > 0:
                used_line_patterns.append(d)
        return used_line_patterns

    def _post_action_update_used_line_patterns(self, doc):
        return_value = res.Result()
        try:
            # find all line patterns of nested families
            nested_family_data = self._find_nested_families_data()
            # get used sub categories from nested data
            nested_family_used_line_patterns = self._get_used_line_patterns(
                nested_family_data
            )
            # update root family data only
            rootFamilyData = self._find_root_family_data()
            # update root processor data as required
            self._update_root_family_data(
                rootFamilyData, nested_family_used_line_patterns
            )
            return_value.update_sep(
                True, "Post Action Update line pattern data successful completed."
            )
        except Exception as e:
            return_value.update_sep(
                False,
                "Post Action Update line pattern data failed with exception: " + str(e),
            )
        return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to image links.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import Autodesk.Revit.DB as rdb

from duHast.Revit.Common import common as com
from duHast.Revit.Common.revit_version import get_revit_version_number

def get_images_types_in_model(doc):
    """
    Gets all image link types and image link instances in a model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A filtered element collector of image types.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = rdb.FilteredElementCollector(doc).OfClass(rdb.ImageType)
    return collector


def get_images_type_ids_in_model(doc):
    """
    Gets all image link type Ids and image link instance ids in a model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of ids representing image types and image instances.
    :rtype: list  Autodesk.Revit.DB.ElementId
    """

    ids = []
    col = get_images_types_in_model(doc)
    ids = com.get_ids_from_element_collector(col)
    return ids


def sort_image_link_types_by_import_or_linked(doc):
    """
    Returns two lists: First one: images linked into model, second one images saved into model from model itself (no external file reference)
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: Two lists of image types and instances (?)
    :rtype: list Autodesk.Revit.DB.ImageType, list Autodesk.Revit.DB.ImageType
    """

    image_link = []
    image_import = []
    collector_image_types = get_images_types_in_model(doc)
    # check revit version to cater for API changes
    revit_version = get_revit_version_number(doc)
    if(revit_version<=2021):
        for im in collector_image_types:
            if im.IsLoadedFromFile():
                image_link.append(im)
            else:
                image_import.append(im)
    else:
        for im in collector_image_types:
            if im.Source == rdb.ImageTypeSource.Link:
                image_link.append(im)
            elif im.Source == rdb.ImageTypeSource.Import:
                image_import.append(im)
    return image_link, image_import


def get_all_image_link_type_linked_in_model(doc):
    """
    Gets all image link types which are links (external referenced) in a model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A list of image types and instances (?)
    :rtype: list Autodesk.Revit.DB.ImageType
    """

    image_links, image_import = sort_image_link_types_by_import_or_linked(doc)
    return image_links


def get_all_image_link_type_imported_in_model(doc):
    """
    Gets all image link types which are imported (not an external reference) in a model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A list of image types and instances (?)
    :rtype: list Autodesk.Revit.DB.ImageType
    """

    image_links, image_import = sort_image_link_types_by_import_or_linked(doc)
    return image_import


def get_image_types_in_model(doc):
    """
    Gets all image types in a model.
    Filters by category.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A filtered element collector of image types.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    return (
        rdb.FilteredElementCollector(doc)
        .OfCategory(rdb.BuiltInCategory.OST_RasterImages)
        .WhereElementIsElementType()
    )


def get_image_instances_in_model(doc):
    """
    Gets all image instances placed in a model.
    Filters by category.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A filtered element collector of image instances.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    return (
        rdb.FilteredElementCollector(doc)
        .OfCategory(rdb.BuiltInCategory.OST_RasterImages)
        .WhereElementIsNotElementType()
    )



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Family Base data analysis module containing functions to find circular family referencing in extracted data.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

A circular reference is when a family A has a family B nested but family B has also family A nested.

Algorithm description:

- read base data file processing list (created by RevitFamilyBaseDataProcessor module)
- extract all root families (no ' :: ' in root path)
- extract all nested families (' :: 'in root path)

- loop over root families
    - loop over nested families 
        - check if root family is in nested family root path
            - yes
            - get all families higher in root path and add to parents prop of root family (if not there already)
            - get all families lower in root path of nested family and add to child property of root family if not already
            -   TODO: Check this theory: families always have the same children...once children are identified there is no need to check over and over again!

- loop over root family 
    -   check whether any family exist in parent and child family
        - YES: found circular reference

"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import threading
import os

from duHast.Revit.Family.Data import family_base_data_utils as rFamBaseDataUtils
from duHast.Utilities.Objects.timer import Timer
from duHast.Utilities.Objects import result as res


def _extract_parent_families(current_parent, tree_path):
    """
    Find the index of the match in the root tree, any entries in the root tree list with a lower index are parents

    Note: Changes current_parent.parent property of the current_parent variable!

    :param current_parent: A tuple containing family root data
    :type current_parent: named tuple rootFamily
    :param tree_path: list of family names describing the nesting tree of a family
    :type tree_path: [str]
    :return: Nothing
    :rtype: None
    """

    index_match = tree_path.index(current_parent.name)
    # double check...it exists and it is not root itself
    if index_match > 0:
        # add all parents
        for i in range(index_match):
            if tree_path[i] not in current_parent.parent:
                current_parent.parent.append(tree_path[i])


def _extract_child_families(current_parent, tree_path):
    """
    Find the index of the match in the root tree, any entries in the root tree list with a lower index are children

    Note: Changes current_parent.child property of the current_parent variable!

    :param current_parent: A tuple containing family root data
    :type current_parent: named tuple rootFamily
    :param tree_path: list of family names describing the nesting tree of a family
    :type tree_path: [str]
    :return: Nothing
    :rtype: None
    """

    index_match = tree_path.index(current_parent.name)
    # double check...it exists and it is not root itself and its not the last item in tree path
    if index_match > 0 and index_match != len(tree_path):
        # add all children
        for i in range(index_match + 1, len(tree_path)):
            if tree_path[i] not in current_parent.child:
                current_parent.child.append(tree_path[i])


def _check_data_blocks_for_over_lap(block_one, block_two):
    """
    Checks whether the root path of families in the first block overlaps with the root path of any family in the second block.
    Overlap is checked from the start of the root path. Any families from block one which are not overlapping any family in\
        block two are returned.

    :param block_one: List of family tuples of type nestedFamily
    :type block_one: [nestedFamily]
    :param block_two: List of family tuples of type nestedFamily
    :type block_two: [nestedFamily]
    :return: List of family tuples of type nestedFamily
    :rtype: [nestedFamily]
    """

    unique_tree_nodes = []
    for fam in block_one:
        match = False
        for fam_up in block_two:
            if " :: ".join(fam_up.rootPath).startswith(" :: ".join(fam.rootPath)):
                match = True
                break
        if match == False:
            unique_tree_nodes.append(fam)
    return unique_tree_nodes


def _cull_data_block(family_base_nested_data_block):
    """
    Sorts family data blocks into a dictionary where key, from 1 onwards, is the level of nesting indicated by number of '::' in root path string.

    After sorting it compares adjacent blocks in the dictionary (key and key + 1) for overlaps in the root path string. Only unique families will be returned.

    :param family_base_nested_data_block: A list containing all nested families belonging to a single root host family.
    :type family_base_nested_data_block: [nestedFamily]
    :return: A list of unique families in terms of root path.
    :rtype: [nestedFamily]
    """

    culled_family_base_nested_data_blocks = []
    data_blocks_by_length = {}
    # build dic by root path length
    # start at 1 because for nesting level ( 1 based rather then 0 based )
    for family in family_base_nested_data_block:
        if len(family.rootPath) - 1 in data_blocks_by_length:
            data_blocks_by_length[len(family.rootPath) - 1].append(family)
        else:
            data_blocks_by_length[len(family.rootPath) - 1] = [family]

    # loop over dictionary and check block entries against next entry up blocks
    for i in range(1, len(data_blocks_by_length) + 1):
        # last block get automatically added
        if i == len(data_blocks_by_length):
            culled_family_base_nested_data_blocks = (
                culled_family_base_nested_data_blocks + data_blocks_by_length[i]
            )
        else:
            # check for matches in next one up
            unique_nodes = _check_data_blocks_for_over_lap(
                data_blocks_by_length[i], data_blocks_by_length[i + 1]
            )
            # only add non overlapping blocks
            culled_family_base_nested_data_blocks = (
                culled_family_base_nested_data_blocks + unique_nodes
            )
    return culled_family_base_nested_data_blocks


def _cull_nested_base_data_blocks(overall_family_base_nested_data):
    """
    Reduce base data families for parent / child finding purposes. Keep the nodes with the root path longest branch only.

    Sample:

    famA :: famB :: famC
    famA :: famB

    The second of the above examples can be culled since the first contains the same information.

    :param overall_family_base_nested_data: A list containing all nested families with the longest nesting levels per branch per host family.
    :type overall_family_base_nested_data: [nestedFamily]
    """

    current_root_fam_name = ""
    family_blocks = []
    block = []
    # read families into blocks
    for nested in overall_family_base_nested_data:
        if nested.rootPath[0] != current_root_fam_name:
            # read family block
            if len(block) > 0:
                family_blocks.append(block)
                # reset block
                block = []
                block.append(nested)
                current_root_fam_name = nested.rootPath[0]
            else:
                block.append(nested)
                current_root_fam_name = nested.rootPath[0]
        else:
            block.append(nested)

    retained_family_base_nested_data = []
    # cull data per block
    for family_block in family_blocks:
        d = _cull_data_block(family_block)
        retained_family_base_nested_data = retained_family_base_nested_data + d

    return retained_family_base_nested_data


def find_parents_and_children(
    overall_family_base_root_data, overall_family_base_nested_data
):
    """
    Loop over all root families and check if they exist in root path of any nested families.
    if so extract families higher up the root path tree as parents and families further down the root path tree as children

    :param overall_family_base_root_data: List of tuples containing root family data.
    :type overall_family_base_root_data: [rootFamily]
    :param overall_family_base_nested_data: List of tuples containing nested family data.
    :type overall_family_base_nested_data: [nestedFamily]

    :return: List of tuples containing root family data.
    :rtype: [rootFamily]
    """

    for i in range(len(overall_family_base_root_data)):
        # print ('checking family :' , i, ' ', overall_family_base_root_data[i].name)
        for nested_fam in overall_family_base_nested_data:
            try:
                # get the index of the match
                index_match = nested_fam.rootPath.index(
                    overall_family_base_root_data[i].name
                )
                if index_match > 0:
                    # print('found ', overall_family_base_root_data[i].name ,' in ', nested_fam.rootPath)
                    _extract_parent_families(
                        overall_family_base_root_data[i], nested_fam.rootPath
                    )

                    _extract_child_families(
                        overall_family_base_root_data[i], nested_fam.rootPath
                    )

                    # print('after: ', overall_family_base_root_data[i].child)
            except:
                pass
    return overall_family_base_root_data


def find_circular_references(overall_family_base_root_data):
    """
    Loops over family data and returns any families which appear in circular references.
    (A family appears in their parent and child collection)

    :param overall_family_base_root_data: List of tuples containing root family data.
    :type overall_family_base_root_data: [rootFamily]

    :return: List of tuples containing root family data.
    :rtype: [rootFamily]
    """

    circular_references = []
    # loop over all families and check whether there are any families in both the parent as well as child collection
    for family in overall_family_base_root_data:
        for parent_family in family.parent:
            if parent_family in family.child:
                circular_references.append(family)
    return circular_references


def check_families_have_circular_references(family_base_data_report_file_path):
    """
    Processes a family base data report and identifies any families which contain circular reference.

    Makes use of multithreading when more then 2 cores are present.

    :param family_base_data_report_file_path: Fully qualified file path to family base data report file.
    :type family_base_data_report_file_path: str

    :return:
        Result class instance.

        - result.status. True if circular referencing file was written successfully, otherwise False.
        - result.message will contain the summary messages of the process including time stamps.
        - result.result empty list

        On exception:

        - result.status (bool) will be False.
        - result.message will contain generic exception message.
        - result.result will be empty

    :rtype: :class:`.Result`
    """

    # set up a timer
    t_process = Timer()
    t_process.start()

    return_value = res.Result()
    # read overall family base data and nested data from file
    (
        overall_family_base_root_data,
        overall_family_base_nested_data,
    ) = rFamBaseDataUtils.read_overall_family_data_list(
        family_base_data_report_file_path
    )
    return_value.append_message(
        "{} Read overall family base data report. {} root entries found and {} nested entries found.".format(
            t_process.stop(),
            len(overall_family_base_root_data),
            len(overall_family_base_nested_data),
        )
    )
    t_process.start()

    before = len(overall_family_base_nested_data)
    # reduce workload by culling not needed nested family data
    overall_family_base_nested_data = _cull_nested_base_data_blocks(
        overall_family_base_nested_data
    )
    return_value.append_message(
        " {} culled nested family base data from : {} to: {} families.".format(
            t_process.stop(), before, len(overall_family_base_nested_data)
        )
    )
    t_process.start()

    # set up some multithreading
    core_count = int(os.environ["NUMBER_OF_PROCESSORS"])
    if core_count > 2:
        return_value.append_message("cores: ".format(core_count))
        # leave some room for other processes
        core_count = core_count - 1
        chunk_size = len(overall_family_base_root_data) / core_count
        threads = []
        # set up threads
        for i in range(core_count):
            t = threading.Thread(
                target=find_parents_and_children,
                args=(
                    overall_family_base_root_data[
                        i * chunk_size : (i + 1) * chunk_size
                    ],
                    overall_family_base_nested_data,
                ),
            )
            threads.append(t)
        # start up threads
        for t in threads:
            t.start()
        # wait for results
        for t in threads:
            t.join()
    else:
        # find parents and children
        overall_family_base_root_data = find_parents_and_children(
            overall_family_base_root_data, overall_family_base_nested_data
        )

    return_value.append_message(
        "{} Populated parents and children properties of: {} root families".format(
            t_process.stop(), len(overall_family_base_root_data)
        )
    )
    t_process.start()

    # identify circular references
    circular_references = find_circular_references(overall_family_base_root_data)
    return_value.append_message(
        "{} Found: {} circular references in families.".format(
            t_process.stop(), len(circular_references)
        )
    )
    if len(circular_references) > 0:
        return_value.result = circular_references
    return return_value
